/**
 * Es la API para conectarse con Pagadito y realizar cobros de forma segura.
 *
 * LICENCIA: Éste código fuente es de uso libre. Su comercialización no está
 * permitida. Toda publicación o mención del mismo, debe ser referenciada a
 * su autor original Pagadito.com.
 *
 * @author      Pagadito.com <soporte@pagadito.com>
 * @copyright   Copyright (c) 2015, Pagadito.com
 * @version     Java 1.6.0
 * @link        https://dev.pagadito.com/index.php?mod=docs&hac=apipg#java
 */

package com.pagadito.apipg;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.StringReader;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.parsers.*;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.JSONValue;
import org.xml.sax.InputSource;
import org.w3c.dom.*;

public class Pagadito {
    
    //*********************************** Atributos
    
    private String uid;
    private String wsk;
    private String apipg;
    private String apipg_sandbox;
    private String format_return;
    private HashMap response;
    private boolean sandbox_mode;
    private String op_connect_key;
    private String op_exec_trans_key;
    private String op_get_status_key;
    private List<HashMap> details;
    private HashMap<String, String> custom_params;
    private String currency;
    private boolean allow_pending_payments;

    //***********************************  Funciones Públicas

    /**
     * Constructor de la clase, el cual inicializa los valores por defecto.
     * @param String uid El identificador del Pagadito Comercio.
     * @param String wsk La clave de acceso.
     */    
    public Pagadito(String uid, String wsk){
        this.uid = uid;
        this.wsk = wsk;
        this.config();
    }
    
    /**
     * Conecta con Pagadito y autentica al Pagadito Comercio.
     * @return boolean
     */
    public boolean connect(){
        HashMap params = new HashMap();
        params.put("operation", this.op_connect_key);
        params.put("uid", this.uid);
        params.put("wsk", this.wsk);
        params.put("format_return", this.format_return);
        try {
            this.response = this.call(params, false);
            if(this.get_rs_code().equals("PG1001")){
                return true;
            }
            else{
                return false;
            }            
        } catch (Exception err){
            Logger.getLogger(Pagadito.class.getName()).log(Level.SEVERE, null, err);
            return false;
        }
    }
    
    /**
     * Solicita el registro de la transacción y redirecciona a la pantalla de
     * cobros de Pagadito. En caso de error devuelve false.
     * @param string ern External Reference Number. Es un número único y obligatorio que identifica una transacción, provisto por el Pagadito Comercio y se utiliza para rastrear las transacciones realizadas por éste.
     * @return boolean
     */
    public boolean exec_trans(String ern){
        if(this.get_rs_code().equals("PG1001")){
            HashMap params = new HashMap();
            params.put("operation", this.op_exec_trans_key);
            params.put("token", this.get_rs_value());
            params.put("ern", ern);
            params.put("amount", this.calc_amount());
            params.put("details", this.encode_details(this.details));
            params.put("custom_params", this.get_encode_custom_params());
            params.put("currency", this.currency);
            params.put("format_return", this.format_return);
            params.put("allow_pending_payments", this.allow_pending_payments);
            this.response = this.call(params, false);
            if(this.get_rs_code().equals("PG1002")){
                return true;
            }
            else{
                return false;
            }
        }
        else{
            return false;
        }
    }

    /**
     * Solicita el estado de una transacción en base a su token.
     * @param string token_trans El identificador de la conexión a consultar.
     * @return boolean
     */
    public boolean get_status(String token_trans){
        if(this.get_rs_code().equals("PG1001")){
            HashMap params = new HashMap();
            params.put("operation", this.op_get_status_key);
            params.put("token", this.get_rs_value());
            params.put("token_trans", token_trans);
            params.put("format_return", this.format_return);
            this.response = this.call(params, true);
            if(this.get_rs_code().equals("PG1003")){
                return true;
            }
            else{
                return false;
            }
        }
        else{
            return false;
        }
    }

    /**
     * Agrega un detalle a la orden de cobro, previo a su ejecución.
     * @param int quantity Define la cantidad del producto.
     * @param string description Define la descripción del producto.
     * @param double price Define el precio del producto en términos de dólares americanos (USD).
     * @param string url_product Define la url de referencia del producto.
     */
    public void add_detail(int quantity, String description, Double price, String url_product){
        HashMap detail = new HashMap();
        detail.put("quantity", Integer.toString(quantity));
        detail.put("description", description);
        detail.put("price", price.toString());
        detail.put("url_product", url_product);
        this.details.add(detail);
    }

    /**
     * Devuelve el código de la respuesta.
     * @return string
     */
    public String get_rs_code(){
        return this.return_attr_response("code");
    }

    /**
     * Devuelve el mensaje de la respuesta.
     * @return string
     */
    public String get_rs_message(){
        return this.return_attr_response("message");
    }

    /**
     * Devuelve el valor de la respuesta.
     * @return object
     */
    public String get_rs_value(){
        return this.return_attr_response("value");
    }

    /**
     * Devuelve la fecha y hora de la respuesta.
     * @return string
     */
    public String get_rs_datetime(){
        return this.return_attr_response("datetime");
    }

    /**
     * Devuelve el estado de la transacción consultada, después de un get_status().
     * @return string
     */
    public String get_rs_status(){
        return this.return_attr_response("status").toString();
    }

    /**
     * Devuelve la referencia de la transacción consultada, después de un get_status().
     * @return string
     */
    public String get_rs_reference(){
        return this.return_attr_response("reference").toString();
    }

    /**
     * Devuelve la fecha y hora de la transacción consultada, después de un get_status().
     * @return string
     */
    public String get_rs_date_trans(){
        return this.return_attr_response("date_trans").toString();
    }

    //*********************************** Funciones Públicas auxiliares

    /**
     * Habilita el modo de pruebas SandBox.
     */
    public void mode_sandbox_on(){
        this.sandbox_mode = true;
    }
    /**
     * Habilita la recepción de pagos preautorizados para la orden de cobro.
     */
    public void enable_pending_payments() {
        this.allow_pending_payments = true;
    }

    public void set_custom_param(String code, String value) {
        this.custom_params.put(code, value);
    }
    /**
     * Cambia el formato de retorno a JSON.
     */
    public void change_format_json(){
        this.format_return = "json";
    }

    /**
     * Cambia el formato de retorno a XML.
     */
    public void change_format_xml(){
        this.format_return = "xml";
    }

    /**
     * Cambia el formato de retorno a PHP.
     */
    public void change_format_php(){
        this.format_return = "php";
    }

    //*********************************** Funciones Privadas

    /**
     * Establece los valores por defecto.
     */
    private void config(){
        this.apipg                  = "https://comercios.pagadito.com/apipg/charges.php";
        this.apipg_sandbox          = "https://sandbox.pagadito.com/comercios/apipg/charges.php";
        this.format_return          = "json";
        this.sandbox_mode           = false;
        this.op_connect_key         = "f3f191ce3326905ff4403bb05b0de150";
        this.op_exec_trans_key      = "41216f8caf94aaa598db137e36d4673e";
        this.op_get_status_key      = "0b50820c65b0de71ce78f6221a5cf876";
        this.details                = new ArrayList<HashMap>();
        this.custom_params          = new HashMap<String, String>();
        this.currency               = "USD";
        this.allow_pending_payments = false;
    }

    /**
     * Devuelve el valor del atributo solicitado.
     * @param string attr Nombre del atributo de la respuesta.
     * @return string
     */
    private String return_attr_response(String attr){
        if(this.response.containsKey(attr)){
            return this.response.get(attr).toString();
        }
        else{
            return "";
        }
    }

    /**
     * Ejecuta una llamada a Pagadito y devuelve la respuesta.
     * @param HashMap params Variables y sus valores a enviarse en la llamada.
     * @param boolean extended Define si es una respuesta extendida la que se
     * recibirá.
     * @return HashMap
     */
    private HashMap call(HashMap params, boolean extended){
        URL url;
        try{
            System.setProperty("https.protocols", "TLSv1.2");
            if(this.sandbox_mode){
                url = new URL(this.apipg_sandbox);
            }
            else{
                url = new URL(this.apipg);
            }
            try {
                URLConnection conn = url.openConnection();                
                conn.setDoOutput(true);
                OutputStreamWriter wr = new OutputStreamWriter(conn.getOutputStream());
                wr.write(this.format_post_vars(params));
                wr.flush();
                BufferedReader rd = new BufferedReader(new InputStreamReader(conn.getInputStream()));
                String respuesta = "";
                for (String line; (line = rd.readLine()) != null;) {
                    respuesta += line;
                }
                wr.close();
                rd.close();
                if(extended){
                    return this.decode_response_extended(respuesta);
                }
                else{
                    return this.decode_response(respuesta);
                }
            } catch (Exception e) {
                Logger.getLogger(Pagadito.class.getName()).log(Level.SEVERE, null, e);
                return null;
            }
        }
        catch(Exception err){
            Logger.getLogger(Pagadito.class.getName()).log(Level.SEVERE, null, err);
            return null;
        }
    }

    /**
     * Devuelve una cadena con el formato válido de los detalles de los productos
     * a enviar en una llamada.
     * @param List<HashMap> details Estructura con los datos de los detalles.
     * @return string
     */
    private String encode_details(List<HashMap> details){
        String encoded_details = "";
        if(this.format_return.equals("xml")){
            HashMap detail = new HashMap();
            encoded_details = "<details>\n";
            for(int i = 0; i < details.size(); i++){
                detail = details.get(i);
                encoded_details += "<detail>\n";
                encoded_details += "<quantity>" + detail.get("quantity") + "</quantity>\n";
                encoded_details += "<description>" + detail.get("description") + "</description>\n";
                encoded_details += "<price>" + detail.get("price") + "</price>\n";
                encoded_details += "<url_product>" + detail.get("url_product") + "</url_product>\n";
                encoded_details += "</detail>\n";
            }
            encoded_details += "</details>";
        }
        else{
            JSONArray array = new JSONArray();
            JSONObject obj = new JSONObject();
            for(int i = 0; i < details.size(); i++){
                HashMap detail = new HashMap();
                detail = details.get(i);
                obj = new JSONObject();
                obj.put("quantity", detail.get("quantity"));
                obj.put("description", detail.get("description"));
                obj.put("price", detail.get("price"));
                obj.put("url_product", detail.get("url_product"));
                array.add(obj);
            }
            encoded_details = array.toString();
        }
        return encoded_details;
    }

    /**
     * Devuelve una cadena con el formato válido de los parámetros personalizados
     * a enviar en una llamada.
     * @param List<HashMap> details Estructura con los datos de los detalles.
     * @return string
     */
    private String get_encode_custom_params(){
        String encoded_params = "";
        if(this.format_return.equals("xml")){
            encoded_params = "<custom_params>\n";
            Set<String> keys = this.custom_params.keySet();
            for (String key : keys) {
                encoded_params += "<"+key+">";
                encoded_params += this.custom_params.get(key);
                encoded_params += "</"+key+">\n";
            }            
            encoded_params += "</custom_params>";
        }
        else{
            JSONObject obj = new JSONObject();;
            Set<String> keys = this.custom_params.keySet();
            for (String key : keys) {
                obj.put(key, this.custom_params.get(key));
            }
            encoded_params = obj.toJSONString();
        }
        return encoded_params;
    }

        
    /**
     * Devuelve una cadena con el formato válido de variables y valores para
     * enviar en una llamada.
     * @param HashMap vars Variables y valores a ser formateados.
     * @return string
     */
    private String format_post_vars(HashMap vars){
        String formatted_vars = "";
        Iterator it = vars.entrySet().iterator();
        while(it.hasNext()){
            Map.Entry e = (Map.Entry)it.next();
            try{
                formatted_vars += URLEncoder.encode(e.getKey().toString(), "UTF-8") + "=" + URLEncoder.encode(e.getValue().toString(), "UTF-8") + "&";
            }
            catch(Exception err){
                
            } 
        }
        formatted_vars = formatted_vars.substring(0, formatted_vars.length() - 1);
        return formatted_vars;
    }

    /**
     * Devuelve un objeto con los datos de la respuesta de Pagadito.
     * @param string response Cadena contenedora de la estructura a ser decodificada.
     * @return HashMap
     */
    private HashMap decode_response(String response){
        HashMap obj_response = new HashMap();
        if(this.format_return.equals("xml")){
            try{
                DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                DocumentBuilder db = dbf.newDocumentBuilder();
                InputSource is = new InputSource();
                is.setCharacterStream(new StringReader(response));
                Document doc = db.parse(is);
                NodeList nodes = doc.getElementsByTagName("response");
                Element element = (Element) nodes.item(0);

                //code
                String code = get_xml_value(element, "code");
                obj_response.put("code", code);
                //message
                String message = get_xml_value(element, "message");
                obj_response.put("message", message);
                //value
                String value = get_xml_value(element, "value");
                obj_response.put("value", value);
                //datetime
                String datetime = get_xml_value(element, "datetime");            
                obj_response.put("datetime", datetime);
            }
            catch(Exception err){
                err.printStackTrace();
            }
        }
        else{
            try{
                Object obj = JSONValue.parse(response);
                JSONObject obj_json = (JSONObject)obj;
                //code
                obj_response.put("code", obj_json.get("code"));
                //message
                obj_response.put("message", obj_json.get("message"));
                //value
                obj_response.put("value", obj_json.get("value"));
                //datetime
                obj_response.put("datetime", obj_json.get("datetime"));
            }
            catch(Exception err){
                err.printStackTrace();
            }
        }
        return obj_response;
    }
    
    /**
     * Devuelve un objeto con los datos de la respuesta de Pagadito.
     * @param string response Cadena contenedora de la estructura a ser decodificada.
     * @return HashMap
     */
    private HashMap decode_response_extended(String response){
        HashMap obj_response = new HashMap();
        if(this.format_return.equals("xml")){
            try{
                DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                DocumentBuilder db = dbf.newDocumentBuilder();
                InputSource is = new InputSource();
                is.setCharacterStream(new StringReader(response));
                Document doc = db.parse(is);
                NodeList nodes = doc.getElementsByTagName("response");
                Element element = (Element) nodes.item(0);

                //code
                String code = get_xml_value(element, "code");
                obj_response.put("code", code);
                //message
                String message = get_xml_value(element, "message");
                obj_response.put("message", message);
                //value
                Element value = get_xml_element(element, "value");
                obj_response.put("value", value);
                //status
                String status = get_xml_value(value, "status");
                obj_response.put("status", status);
                //status
                String reference = get_xml_value(value, "reference");
                obj_response.put("reference", reference);
                //status
                String date_trans = get_xml_value(value, "date_trans");
                obj_response.put("date_trans", date_trans);
                //datetime
                String datetime = get_xml_value(element, "datetime");            
                obj_response.put("datetime", datetime);
            }
            catch(Exception err){
                err.printStackTrace();
            }
        }
        else{ 
            try{
                Object obj = JSONValue.parse(response);
                JSONObject obj_json = (JSONObject)obj;
                //code
                obj_response.put("code", obj_json.get("code"));
                //message
                obj_response.put("message", obj_json.get("message"));
                //value
                obj_response.put("value", obj_json.get("value"));
                //datetime
                obj_response.put("datetime", obj_json.get("datetime"));
                
                obj = JSONValue.parse(obj_response.get("value").toString());
                obj_json = (JSONObject)obj;
                //status
                obj_response.put("status", obj_json.get("status"));
                //reference
                obj_response.put("reference", obj_json.get("reference"));
                //date_trans
                obj_response.put("date_trans", obj_json.get("date_trans"));
            }
            catch(Exception err){
                err.printStackTrace();
            }
        }
        return obj_response;
    }
    
    /**
     * Devuelve el valor del ítem solicitado.
     * @param Element element Objeto con datos provenientes de un XML.
     * @param String item Nombre del ítem solicitado.
     * @return String
     */
    private String get_xml_value(Element element, String item){
        NodeList name = element.getElementsByTagName(item);
        Element line = (Element) name.item(0);
        Node child = line.getFirstChild();
        if (child instanceof CharacterData) {
            CharacterData cd = (CharacterData) child;
            String value = cd.getData();
            return value;
        }
        else{
            return "";
        }
    }
    
    /**
     * Devuelve el objeto del ítem solicitado.
     * @param Element element Objeto con datos provenientes de un XML.
     * @param String item Nombre del ítem solicitado.
     * @return Element
     */
    public Element get_xml_element(Element element, String item){
        NodeList name = element.getElementsByTagName(item);
        Element line = (Element) name.item(0);
        return line;
    }
   
    /**
     * Devuelve la sumatoria de los productos entre cantidad y precio de todos
     * los detalles de la transacción.
     * @return double
     */
    private double calc_amount(){
        double amount = 0;
        for(int i = 0; i < this.details.size(); i++){
            amount += Double.parseDouble(this.details.get(i).get("quantity").toString()) * Double.parseDouble(this.details.get(i).get("price").toString());
        }
        return amount;
    }
    
}
